@@ -1,11 +1,13 @@
 from time import sleep
 from parfor import parfor
+import numpy as np
 
 
 def my_fun(*args, **kwargs):
-    @parfor(range(10), (3,))
+    @parfor(range(10), (3,), length=10)
     def fun(i, a):
-        sleep(1)
+        sleep(np.random.randint(1, 60))
+        assert False
         return a * i ** 2
 
     return fun
